local addon_ToCName, addon = ...;

LFG_TIMER_CLONE_FRAME_HEIGHT = 153;
LFG_TIMER_CLONE_FRAME_WIDTH = 300;

_G.LFGTimer = LibStub("AceAddon-3.0"):NewAddon(addon, addon_ToCName, "AceEvent-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("LFGTimer");
local LSM = LibStub("LibSharedMedia-3.0");

LSM:Register("Background", "Blizzard Chat Bubble", "Interface\\Tooltips\\ChatBubble-Background");

local db;
local defaults = {
	profile = {
		ShowFrame = true,
		FrameType = "custom",
		Frame = {
			Tooltip = {
				Data = {
					Tank = true,
					Healer = true,
					DPS = true,
					Self = true,
					Party = true,
					Type = true,
					Found = {
						Total = true,
						Tank = true,
						Healer = true,
						DPS = true,
					},
				},
				Backdrop = {
					Background = "Blizzard Chat Bubble",
					Border = "Blizzard Chat Bubble",
					BackgroundColor = {
						r = 0,
						g = 0,
						b = 0,
						a = 1,
					},
					BorderColor = {
						r = 1,
						g = 1,
						b = 1,
						a = 1,
					},
					Inset = 5,
					Tile = false,
					TileSize = 32,
					EdgeSize = 24,
				},
				Text = {
					Font = LSM:GetDefault("font"),
					Size = 12,
					Outline = " ",
					Color = {
						r = 1,
						g = 1,
						b = 1,
						a = 1,
					},
				},
			},
			ShowWhen = {
				Not = true,
				Acquire = true,
				Wait = true,
				Party = true,
				Found = true,
			},
			Position = {
				AnchorAtSelf = "CENTER",
				AnchoredTo = "UIParent",
				AnchorAtOther = "CENTER",
				AnchorXOffset = 0,
				AnchorYOffset = 0,
			},
		},
	},
};

--! @todo Update .Options to reflect the new DB structure
addon.Options = {
	type = "group",
	order = 1,
	name = "LFG Timer",
	get = function(info)
		return db[ info[#info] ];
	end,
	set = function(info, value)
		db[ info[#info] ] = value;
		addon:UpdateFrameArt();
		addon:UpdateFrameShown();
	end,
	args = {
		General = {
			type = 'group',
			order = 1,
			name = L["Frame options"],
			hidden = function(info)
				return not db.ShowFrame;
			end,
			childGroups = 'tab',
			args = {
				ShowFrame = {
					type = 'toggle',
					order = 1,
					name = L["Show frame"],
					desc = L["Show an LFD search status tooltip"],
					hidden = false,
				},
				FrameType = {
					type = 'select',
					order = 2,
					name = L["Frame type"],
					desc = L["Type of search status frame to show.\n\n|cffffd200Search status clone|r: Clone of the Blizzard LFD Search Status frame.\n\n|cffffd200Customizable tooltip|r: Customizable tooltip that can show detailed information about LFD queues."],
					values = {
						clone = L["Search status clone"],
						custom = L["Customizable tooltip"],
					},
				},
				ShowWhen = {
					type = 'group',
					order = 3,
					name = L["Show when..."],
					guiInline = true,
					get = function(info)
						return db.Frame.ShowWhen[ info[#info] ];
					end,
					set = function(info, value)
						db.Frame.ShowWhen[ info[#info] ] = value;
					end,
					args = {
						Not = {
							type = "toggle",
							order = 1,
							name = L["Not queued"],
							desc = L["Show the frame when you are not in an LFG queue."],
							width = 'full',
						},
						Acquire = {
							type = "toggle",
							order = 2,
							name = L["Acquiring data"],
							desc = L["Show the frame while acquring LFG queue data."],
							width = 'full',
						},
						Wait = {
							type = "toggle",
							order = 3,
							name = L["Waiting for a group"],
							desc = L["Show the frame while waiting in the queue."],
							width = 'full',
						},
						Party = {
							type = "toggle",
							order = 4,
							name = L["In a party"],
							desc = L["Show the frame while in an LFG party."],
							width = 'full',
						},
						Found = {
							type = "toggle",
							order = 5,
							name = L["Group found"],
							desc = L["Show the frame when a group is found."],
							width = 'full',
						},
					},
				},
			},
		},
		Tooltip = {
			type = 'group',
			order = 2,
			name = L["Customizable tooltip options"],
			childGroups = 'tab',
			get = function(info)
				return db.Frame.Tooltip[ info[#info] ];
			end,
			set = function(info, value)
				db.Frame.Tooltip[ info[#info] ] = value;
				addon:UpdateFrameArt();
			end,
			disabled = function()
				return db.FrameType ~= 'custom';
			end,
			args = {
				__desc = {
					type = 'description',
					order = 0,
					name = L["These options are only available when using the Customizable tooltip frame."],
					hidden = function()
						return db.ShowFrame and db.FrameType == 'custom';
					end,
					disabled = false,
				},
				Data = {
					type = 'group',
					order = 1,
					name = L["Queue data"],
					desc = L["Tooltip information options"],
					get = function(info)
						return db.Frame.Tooltip.Data[ info[#info] ];
					end,
					set = function(info, value)
						db.Frame.Tooltip.Data[ info[#info] ] = value;
					end,
					args = {
						Type = {
							type = 'toggle',
							order = 1,
							name = L["Dungeon queue type"],
							desc = L["Show the type of dungeon queue you are waiting in."],
							width = 'double',
						},
						Tank = {
							type = 'toggle',
							order = 2,
							name = L["Tank queue times"],
							desc = L["Show the estimated queue times for Tanks."],
							width = 'double',
						},
						Healer = {
							type = 'toggle',
							order = 3,
							name = L["Healer queue times"],
							desc = L["Show the estimated queue times for Healers."],
							width = 'double',
						},
						DPS = {
							type = 'toggle',
							order = 4,
							name = L["DPS queue times"],
							desc = L["Show the estimated queue times for Damage dealers."],
							width = 'double',
						},
						Self = {
							type = 'toggle',
							order = 5,
							name = L["Personal queue time"],
							desc = L["Show the estimated queue time for your current group."],
							width = 'double',
						},
						Party = {
							type = 'toggle',
							order = 6,
							name = L["Number of party members found"],
							desc = L["Show information about the players and roles found for the current group."],
							width = 'double',
						},
						Found = {
							type = 'multiselect',
							order = 7,
							name = L["Party members found information"],
							desc = L["Toggle display of total number of players, tanks, healers and damagers found."],
							width = 'double',
							disabled = function()
								return not db.Frame.Tooltip.Data.Party;
							end,
							get = function(info, key)
								return db.Frame.Tooltip.Data.Found[key];
							end,
							set = function(info, key, value)
								db.Frame.Tooltip.Data.Found[key] = value;
							end,
							values = {
								Total = L["Total number of players"],
								Tank = L["Tank"],
								Healer = L["Healer"],
								DPS = L["DPS"],
							},
						},
					},
				},
				Backdrop = {
					type = 'group',
					order = 2,
					name = L["Backdrop"],
					desc = L["Tooltip background and border options"],
					get = function(info)
						if info.type == 'color' then
							local t = db.Frame.Tooltip.Backdrop[ info[#info] ];
							return t.r, t.g, t.b, t.a;
						else
							return db.Frame.Tooltip.Backdrop[ info[#info] ];
						end
					end,
					set = function(info, ...)
						if info.type == 'color' then
							local t = db.Frame.Tooltip.Backdrop[ info[#info] ];
							t.r, t.g, t.b, t.a = ...;
						else
							db.Frame.Tooltip.Backdrop[ info[#info] ] = ...;
						end
						addon:UpdateFrameArt();
					end,
					args = {
						Background = {
							type = 'select',
							order = 1,
							name = L["Background texture"],
							desc = L["Texture used as the tooltip's background."],
							dialogControl = 'LSM30_Background',
							values = AceGUIWidgetLSMlists.background,
							width = 'full',
						},
						BackgroundColor = {
							type = 'color',
							order = 2,
							name = L["Background texture color"],
							desc = L["Color used to tint the tooltip's background texture.\n|cffffd200Note|r: Darker textures are less effected by this option."],
							width = 'full',
						},
						Tile = {
							type = 'toggle',
							order = 3,
							name = L["Tile the background texture"],
							desc = L["Repeat the background texture horizontally and vertically."],
							width = 'full',
						},
						TileSize = {
							type = 'range',
							order = 4,
							name = L["Background tile size"],
							desc = L["Size of the blocks used when tiling the background texture."],
							width = 'full',
							min = 2,
							max = 64,
							step = 1,
							bigStep = 2,
							disabled = function(info)
								return not db.Frame.Tooltip.Backdrop.Tile;
							end,
						},
						Inset = {
							type = 'range',
							order = 5,
							name = L["Background inset"],
							desc = L["Distance between the border and the background texture."],
							width = 'full',
							min = 0,
							max = 32,
							step = 1,
							bigStep = 2,
						},
						Border = {
							type = 'select',
							order = 7,
							name = L["Border texture"],
							desc = L["Texture used as the tooltip's border."],
							dialogControl = 'LSM30_Border',
							values = AceGUIWidgetLSMlists.border,
							width = 'full',
						},
						BorderColor = {
							type = 'color',
							order = 8,
							name = L["Border texture color"],
							desc = L["Color used to tint the tooltip's border texture.\n|cffffd200Note|r: Darker textures are less effected by this option."],
							width = 'full',
						},
						EdgeSize = {
							type = 'range',
							order = 9,
							name = L["Border size"],
							desc = L["Thickness of the border texture."],
							width = 'full',
							min = 2,
							max = 32,
							step = 1,
							bigStep = 2,
						},
					},
				},
				Text = {
					type = 'group',
					order = 3,
					name = L["Text"],
					desc = L["Text appearance options"],
					get = function(info)
						if info.type == 'color' then
							local t = db.Frame.Tooltip.Text[ info[#info] ];
							return t.r, t.g, t.b, t.a;
						else
							return db.Frame.Tooltip.Text[ info[#info] ];
						end
					end,
					set = function(info, ...)
						if info.type == 'color' then
							local t = db.Frame.Tooltip.Text[ info[#info] ];
							t.r, t.g, t.b, t.a = ...;
						else
							db.Frame.Tooltip.Text[ info[#info] ] = ...;
						end
					end,
					args = {
						Font = {
							type = 'select',
							order = 1,
							name = L["Font"],
							desc = L["Font used for the tooltip text."],
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							width = 'full',
						},
						Size = {
							type = 'range',
							order = 2,
							name = L["Text size"],
							desc = L["Font size used for the tooltip text."],
							width = 'full',
							min = 4,
							max = 20,
							step = 1,
						},
						Outline = {
							type = 'select',
							order = 3,
							name = L["Text outline"],
							desc = L["Add an outline to the tooltip text."],
							width = 'full',
							values = {
								[" "] = L["No outline"],
								["OUTLINE"] = L["Thin outline"],
								["THICKOUTLINE"] = L["Thick outline"],
							},
						},
						Color = {
							type = 'color',
							order = 4,
							name = L["Text color"],
							desc = L["Color used for the tooltip text."],
							width = 'full',
						},
					},
				},
			},
		},
	},
};

function addon:OnInitialize()
	self.Vars = LibStub("AceDB-3.0"):New("LFGTimerDB", defaults);
	self.Vars.RegisterCallback(self, "OnProfileChanged", "UpdateDB");
	self.Vars.RegisterCallback(self, "OnProfileCopied", "UpdateDB");
	self.Vars.RegisterCallback(self, "OnProfileReset", "UpdateDB");
	
	db = self.Vars.profile;
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("LFGTimer", self.Options);
	
	local lib = LibStub("AceConfigDialog-3.0");
	self.OptionsFrames = {};
	
	self.Options.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.Vars);
	
	self.OptionsFrames.General = lib:AddToBlizOptions("LFGTimer", "LFG Timer", nil, "General");
	self.OptionsFrames.Tooltip = lib:AddToBlizOptions("LFGTimer", L["Tooltip options"], "LFG Timer", "Tooltip");
	self.OptionsFrames.Profiles = lib:AddToBlizOptions("LFGTimer", L["Profiles"], "LFG Timer", "Profiles");
	
	self.LFGData = {};
	self.LFGMode = {};
end

function addon:OnEnable()
	self:RegisterEvent("LFG_QUEUE_STATUS_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_SHOW", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_FAILED", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_SUCCEEDED", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_CHECK_SHOW", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_CHECK_HIDE", "UpdateFrameShown");
	self:RegisterEvent("LFG_BOOT_PROPOSAL_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("VOTE_KICK_REASON_NEEDED", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_UPDATE_RANDOM_INFO", "UpdateFrameShown");
	self:RegisterEvent("LFG_OPEN_FROM_GOSSIP", "UpdateFrameShown");
	self:RegisterEvent("GOSSIP_CLOSED", "UpdateFrameShown");
end

function addon:OnDisable()
	self:UnregisterAllEvents();
end

function addon:UpdateDB(event, database)
	db = database.profile;
	addon:UpdateFrameArt();
	addon:UpdateFrameShown();
end

function addon:UpdateQueueData()
	for k, v in pairs(self.LFGData) do
		self.LFGData[k] = nil;
	end
	
	for k, v in pairs(self.LFGMode) do
		self.LFGMode[k] = nil;
	end
	
	local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17;
	
	arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17 = GetLFGQueueStats();
	
--	hasData,  leaderNeeds, tankNeeds, healerNeeds, dpsNeeds, totalTanks, totalHealers, totalDPS, instanceType, instanceSubType, instanceName, averageWait, tankWait, healerWait, damageWait, myWait, queuedTime = GetLFGQueueStats();

	self.LFGData.HaveData = arg1;
	self.LFGData.NeedLeaders = arg2 or 0;
	self.LFGData.NeedTanks = arg3 or 0;
	self.LFGData.NeedHealers = arg4 or 0;
	self.LFGData.NeedDPS = arg5 or 0;
	self.LFGData.TotalTanks = arg6 or 1; -- Failsafe counts for missing dungeon info
	self.LFGData.TotalHealers = arg7 or 1;
	self.LFGData.TotalDPS = arg8 or 3;
	self.LFGData.QueueType = arg9;
	self.LFGData.QueueSubType = arg10;
	self.LFGData.QueueName = arg11;
	self.LFGData.AverageWait = arg12;
	self.LFGData.TankWait = arg13;
	self.LFGData.HealerWait = arg14;
	self.LFGData.DPSWait = arg15;
	self.LFGData.MyWait = arg16;
	self.LFGData.QueueTime = arg17;
	
	arg1, arg2 = GetLFGMode();
	
	self.LFGMode.Mode = arg1;
	self.LFGMode.SubMode = arg2;
	
	return self.LFGData, self.LFGMode;
end

function addon:UpdateFrameShown()
	if self.CloneFrame then
		self.CloneFrame:Hide();
	end
	if self.CustomFrame then
		self.CustomFrame:Hide();
	end
	
	if self:ShouldShowFrame() then
		self:GetFrame():Show();
	end
end

function addon:ShouldShowFrame()
	local lfgData, lfgMode = self:UpdateQueueData();
	local ret = false;
	
	if not db.ShowFrame then
		ret = false;
	elseif (not lfgMode.Mode or lfgMode.Mode == "abandonedInDungeon") and not lfgData.HaveData then -- not queud
		ret = db.Frame.ShowWhen.Not;
	elseif lfgMode.Mode == "proposal" then -- queue popped, waiting to join
		ret = db.Frame.ShowWhen.Found;
	elseif (lfgMode.Mode == "queued" or lfgMode.Mode == "listed" or lfgMode.Mode == "rolecheck") and not lfgData.HaveData then -- queued, waiting on data
		ret = db.Frame.ShowWhen.Acquire;
	elseif lfgMode.Mode == "lfgparty" then -- in an instance
		ret = db.Frame.ShowWhen.Party;
	elseif lfgMode.Mode == "queued" and lfgData.HaveData then -- queued, have data
		ret = db.Frame.ShowWhen.Wait;
	else
		ret = false; -- wtf is going on with LFG, hide the frame
	end
	
	return ret;
end

function addon:GetFrame()
	if db.FrameType == "clone" then
		return self:SetupCloneFrame();
	elseif db.FrameType == "custom" then
		return self:SetupCustomFrame();
	end
end

function addon:SetAnchor(frame)
	local anchor = db.Frame.Position;
	frame:ClearAllPoints();
	frame:SetPoint(anchor.AnchorAtSelf, anchor.AnchoredTo, anchor.AnchorAtOther, anchor.AnchorXOffset, anchor.AnchorYOffset);
end

function addon:UpdateFramePosition(frame)
	local anchor = db.Frame.Position;
	local point, to, at, x, y = frame:GetPoint(1);
	anchor.AnchorAtSelf = point;
	anchor.AnchorAtOther = at;
	anchor.AnchorXOffset = x;
	anchor.AnchorYOffset = y;
end

function addon:SetupCloneFrame()
	-- Create clone frame
	local frame, f; -- Declare variables
	
	frame = CreateFrame("Frame", "LFGTimerSearchStatusClone", UIParent);
	
	frame:SetFrameStrata("TOOLTIP");
	frame:SetMovable(true);
	frame:SetClampedToScreen(true);
	
	frame:SetHeight(153);
	frame:SetWidth(300);
	
	frame:ClearAllPoints();
	frame:SetPoint("CENTER");
	
	frame.Backdrop = {
		bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
		edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {
			left = 5,
			right = 5,
			top = 5,
			bottom = 5,
		},
	};
	
	frame:SetBackdrop(frame.Backdrop);
	
	f = frame:CreateFontString("$parentTitle", "ARTWORK", "GameFontHighlight");
	f:SetText(ASSEMBLING_GROUP);
	f:SetPoint("TOP", 0, -15);
	frame.Title = f;
	
	f = frame:CreateFontString("$parentQueueStatistics", "ARTWORK", "GameFontNormal");
	f:SetPoint("BOTTOM", 0, 38);
	f:SetJustifyH("CENTER");
	f:SetJustifyV("CENTER");
	frame.Stats = f;
	
	f = frame:CreateFontString("$parentLookingFor", "ARTWORK", "GameFontHighlight");
	f:SetText(SEARCHING_FOR_GROUPS_NEEDS);
	f:SetPoint("BOTTOM", 0, 14);
	f:SetJustifyH("CENTER");
	frame.LookingFor = f;
	
	f = frame:CreateTexture("$parentRoleIcon1", "ARTWORK", "LFGSearchStatusRoleIconTemplate");
	f:ClearAllPoints();
	f:SetPoint("LEFT", frame.LookingFor, "RIGHT", 2, 0);
	f:SetHeight(25);
	f:SetWidth(25);
	frame.RoleIcon1 = f;
	
	f = frame:CreateTexture("$parentRoleIcon2", "ARTWORK", "LFGSearchStatusRoleIconTemplate");
	f:ClearAllPoints();
	f:SetPoint("LEFT", frame.RoleIcon1, "RIGHT", 2, 0);
	f:SetHeight(25);
	f:SetWidth(25);
	frame.RoleIcon2 = f;
	
	f = frame:CreateTexture("$parentRoleIcon3", "ARTWORK", "LFGSearchStatusRoleIconTemplate");
	f:ClearAllPoints();
	f:SetPoint("LEFT", frame.RoleIcon2, "RIGHT", 2, 0);
	f:SetHeight(25);
	f:SetWidth(25);
	frame.RoleIcon3 = f;
	
	frame.RoleIcons = {
		frame.RoleIcon1,
		frame.RoleIcon2,
		frame.RoleIcon3,
	};
	
	f = CreateFrame("Button", "$parentDragBar", frame);
	f:SetPoint("TOPLEFT", 10, -5);
	f:SetPoint("RIGHT", -10, 0);
	f:SetHeight(8);
	f:SetHighlightTexture([[Interface\FriendsFrame\UI-FriendsFrame-HighlightBar]]);
	f:SetScript("OnMouseDown", function(self, button)
		frame:StartMoving();
	end);
	f:SetScript("OnMouseUp", function(self, button)
		frame:StopMovingOrSizing();
		addon:UpdateFramePosition(frame);
	end);
	frame.Drag = f;
	
	f = CreateFrame("Frame", "$parentRoleDisplay", frame);
	f:SetPoint("TOPLEFT", 25, -40);
	f:SetHeight(1);
	f:SetWidth(1);
	frame.RoleDisplay = f;
	
	f = CreateFrame("Frame", "$parentHealers", frame.RoleDisplay, "LFGSearchStatusRoleWithCountTemplate");
	f:SetPoint("TOP", frame, "TOP", 0, -40);
	f.texture:SetTexCoord(GetTexCoordsForRole("HEALER"));
	frame.Healers = f;
	
	f = CreateFrame("Frame", "$parentTanks", frame.RoleDisplay, "LFGSearchStatusRoleWithCountTemplate");
	f:SetPoint("RIGHT", frame.Healers, "LEFT", -33, 0);
	f.texture:SetTexCoord(GetTexCoordsForRole("TANK"));
	frame.Tanks = f;
	
	f = CreateFrame("Frame", "$parentDPS", frame.RoleDisplay, "LFGSearchStatusRoleWithCountTemplate");
	f:SetPoint("LEFT", frame.Healers, "RIGHT", 33, 0);
	f.texture:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
	frame.DPS = f;
	
	frame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	frame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	
	for k, v in pairs(addon.CloneFrameFuncs) do
		frame[k] = v;
	end
	
	frame:SetScript("OnShow", frame.OnShow);
	frame:SetScript("OnHide", frame.OnHide);
	
	self:SetAnchor(frame);
	
	self.CloneFrame = frame;
	frame:Hide();
	
	self.SetupCloneFrame = function(self)
		return self.CloneFrame;
	end -- Overwrite the setup function to just return the frame
	
	return frame;
end

function addon:SetupCustomFrame()
	-- Create custom data frame
	local frame, f;

	frame = CreateFrame("Frame", "LFGTimerCustomDataFrame", UIParent);

	frame:SetFrameStrata("MEDIUM");
	frame:SetMovable(true);
	frame:SetUserPlaced(true);
	frame:SetClampedToScreen(true);

	frame:ClearAllPoints();
	frame:SetPoint("CENTER");
	frame:SetHeight(150);
	frame:SetWidth(150);

	self:SetAnchor(frame);

	f = CreateFrame("Button", "$parentDragBar", frame);
	f:SetPoint("TOP", 0, -5);
	f:SetPoint("LEFT", 10, 0);
	f:SetPoint("RIGHT", -10, 0);
	f:SetHeight(8);
	f:SetHighlightTexture([[Interface\FriendsFrame\UI-FriendsFrame-HighlightBar]]);
	f:SetScript("OnMouseDown", function(self, button)
		frame:StartMoving();
	end);
	f:SetScript("OnMouseUp", function(self, button)
		frame:StopMovingOrSizing();
		addon:UpdateFramePosition(frame);
	end);
	frame.Drag = f;

	f = CreateFrame("Frame", "$parentContainer", frame);
	f:SetPoint("TOP", frame.Drag, "BOTTOM", 0, -5);
	f:SetPoint("LEFT", 5, 0);
	f:SetPoint("RIGHT", -5, 0);
	f:SetPoint("BOTTOM", 0, 5);
	frame.Container = f;

	frame.Lines = {};
	frame.TextFrames = {};
	frame.NumTextFrames = 0;
	frame.Backdrop = {
		insets = {},
	};

	for k, v in pairs(addon.CustomFrameFuncs) do
		frame[k] = v;
	end

	frame:SetScript("OnShow", frame.OnShow);
	frame:SetScript("OnHide", frame.OnHide);

	frame:Hide();

	frame:UpdateFrameArt();

	self.CustomFrame = frame;
	
	self.SetupCustomFrame = function(self)
		return self.CustomFrame;
	end -- Overwrite the setup function to just return the frame
	
	return frame;
end

function addon:UpdateFrameArt()
	if self.CustomFrame then
		self.CustomFrame:UpdateFrameArt();
	end
end

addon.CloneFrameFuncs = {
	OnShow = function(frame)
		frame:SetScript("OnUpdate", frame.Update);
	end,
	OnHide = function(frame)
		frame:SetScript("OnUpdate", nil);
	end,
	Update = function(frame)
		local lfgData, lfgMode = addon:UpdateQueueData();
		local height = LFG_TIMER_CLONE_FRAME_HEIGHT; -- Debug vars to adjust window dimensions
		local width = LFG_TIMER_CLONE_FRAME_WIDTH;
	
		frame:SetHeight(height);
		
		frame:UpdateRoleIcons();
		frame:UpdateRolesFound();
		
		if (not lfgMode.Mode or lfgMode.Mode == "abandonedInDungeon") and not lfgData.HaveData then
			frame.Stats:SetText(L["You are not in an LFG queue"]);
		elseif lfgMode.Mode == "proposal" then
			frame.Stats:SetText(L["Found a group"]);
			if lfgMode.SubMode == "accepted" then
				frame.Stats:SetText(L["Found a group"] .. "\n" .. L["Waiting on group members"]);
			end
		elseif (lfgMode.Mode == "queued" or lfgMode.Mode == "listed" or lfgMode.Mode == "rolecheck") and not lfgData.HaveData then
			frame.Stats:SetText(L["Acquiring LFG queue data..."]);
		elseif lfgMode.Mode == "lfgparty" then
			frame.Stats:SetText(L["You are in an LFG party"]);
		else
			local elapsedTime = GetTime() - lfgData.QueueTime;
			local text = string.format("%s\n%s", LFG_STATISTIC_AVERAGE_WAIT, TIME_IN_QUEUE);
			frame.Stats:SetFormattedText(text, lfgData.MyWait == -1 and TIME_UNKONWN or SecondsToTime(lfgData.MyWait, false, false, 1), (elapsedTime >= 60) and SecondsToTime(elapsedTime) or LESS_THAN_ONE_MINUTE);
		end
		
		height = height + frame.Stats:GetStringHeight();
		width = math.max(width, frame.Stats:GetStringWidth() + 10);
		
		frame:SetHeight(height);
		frame:SetWidth(width);
	end,
	UpdateRoleIcon = function(icon, need)
		if need then
			SetDesaturation(icon.texture, true);
			icon.cover:Show();
		else
			SetDesaturation(icon.texture, false);
			icon.cover:Hide();
		end
	end,
	UpdateRolesFound = function(frame)
		local lfgData, lfgMode = addon.LFGData, addon.LFGMode;
		if lfgData.HaveData then
			local foundTanks, foundHealers, foundDPS = lfgData.TotalTanks - lfgData.NeedTanks, lfgData.TotalHealers - lfgData.NeedHealers, lfgData.TotalDPS - lfgData.NeedDPS;
			
			frame.Tanks.count:SetFormattedText(PLAYERS_FOUND_OUT_OF_MAX, foundTanks, lfgData.TotalTanks);
			frame.Healers.count:SetFormattedText(PLAYERS_FOUND_OUT_OF_MAX, foundHealers, lfgData.TotalHealers);
			frame.DPS.count:SetFormattedText(PLAYERS_FOUND_OUT_OF_MAX, foundDPS, lfgData.TotalDPS);
			
			frame.Tanks.count:Show();
			frame.Healers.count:Show();
			frame.DPS.count:Show();
			
			frame.UpdateRoleIcon(frame.Tanks, foundTanks ~= lfgData.TotalTanks);
			frame.UpdateRoleIcon(frame.Healers, foundHealers ~= lfgData.TotalHealers);
			frame.UpdateRoleIcon(frame.DPS, foundDPS ~= lfgData.TotalDPS);
		else
			frame.Tanks.count:Hide();
			frame.Healers.count:Hide();
			frame.DPS.count:Hide();
			
			frame.UpdateRoleIcon(frame.Tanks, true);
			frame.UpdateRoleIcon(frame.Healers, true);
			frame.UpdateRoleIcon(frame.DPS, true);
		end
	end,
	UpdateRoleIcons = function(frame)
		local UpdateRoleIcon = addon.CloneFrame.UpdateRoleIcon;
		local leader, tank, healer, damage = GetLFGRoles();
		local icon, i;
		
		i = 1;
		if tank then
			icon = frame["RoleIcon" .. i];
			icon:SetTexCoord(GetTexCoordsForRole("TANK"));
			icon:Show();
			i = i + 1;
		end
		
		if healer then
			icon = frame["RoleIcon" .. i];
			icon:SetTexCoord(GetTexCoordsForRole("HEALER"));
			icon:Show();
			i = i + 1;
		end
		
		if damage then
			icon = frame["RoleIcon" .. i];
			icon:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
			icon:Show();
			i = i + 1;
		end
		
		local extraWidth = 27 * (i - 1);
		while i <= 3 do
			icon = frame["RoleIcon" .. i];
			icon:Hide();
			i = i + 1;
		end
		frame.LookingFor:SetPoint("BOTTOM", -extraWidth / 2, 14);
	end,
	SetQueueMode = function(frame, mode)
		for k, v in pairs(frame.QueueModes) do
			v:Hide();
		end
		
		local f = frame.QueueModes[mode];
		
		if not f then
			return;
		end
		
		f:Show();
		
		frame.QueueMode = mode;
	end,
};

addon.CustomFrameFuncs = {
	OnShow = function(frame)
		frame:SetScript("OnUpdate", frame.UpdateData);
	end,
	OnHide = function(frame)
		frame:SetScript("OnUpdate", nil);
	end,
	UpdateFrameArt = function(frame)
		local inset = db.Frame.Tooltip.Backdrop.Inset;
		local backdrop = frame.Backdrop;
		
		backdrop.bgFile = LSM:Fetch("background", db.Frame.Tooltip.Backdrop.Background);
		backdrop.edgeFile = LSM:Fetch("border", db.Frame.Tooltip.Backdrop.Border);
		backdrop.tile = db.Frame.Tooltip.Backdrop.Tile;
		backdrop.tileSize = db.Frame.Tooltip.Backdrop.TileSize;
		backdrop.edgeSize = db.Frame.Tooltip.Backdrop.EdgeSize;
		
		backdrop.insets.left = inset;
		backdrop.insets.right = inset;
		backdrop.insets.top = inset;
		backdrop.insets.bottom = inset;
		
		local bgColor, borderColor = db.Frame.Tooltip.Backdrop.BackgroundColor, db.Frame.Tooltip.Backdrop.BorderColor;
		
		frame:SetBackdrop(backdrop);
		frame:SetBackdropColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);
		frame:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, borderColor.a);
		
		frame.Drag:SetPoint("TOP", 0, -inset);
		frame.Drag:SetPoint("LEFT", inset, 0);
		frame.Drag:SetPoint("RIGHT", -inset, 0);
		
		frame.Container:ClearAllPoints();
		frame.Container:SetPoint("TOP", frame.Drag, "BOTTOM", 0, -5);
		frame.Container:SetPoint("BOTTOM", 0, inset);
		frame.Container:SetPoint("LEFT", inset, 0);
		frame.Container:SetPoint("RIGHT", -inset, 0);
		
		frame:UpdateSize();
	end,
	ClearLines = function(frame)
		for k, v in pairs(frame.Lines) do
			frame.Lines[k] = nil;
			v:Hide();
			v.Text:SetText("Empty");
			frame.TextFrames[v] = true;
		end
	end,
	GetTextFrame = function(frame)
		local f = next(frame.TextFrames);
		if not f then
			frame.NumTextFrames = frame.NumTextFrames + 1;
			f = CreateFrame("Frame", "LFGTimerTextFrame" .. frame.NumTextFrames, frame.Container, "LFGTimerTextFrameTemplate");
		else
			frame.TextFrames[f] = nil;
		end
		
		local font, size, flags, color = (LSM:Fetch("font", db.Frame.Tooltip.Text.Font)), db.Frame.Tooltip.Text.Size, db.Frame.Tooltip.Text.Outline, db.Frame.Tooltip.Text.Color;
		
		f.Text:SetText("Initialized");
		f.Text:Show();
		f.Text:SetFont(font, size, flags);
		f.Text:SetTextColor(color.r, color.g, color.b, color.a);
		return f;
	end,
	AddLine = function(frame, text)
		local f = frame:GetTextFrame();
		f.Text:SetText(text);
		table.insert(frame.Lines, f);
	end,
	UpdateSize = function(frame)
		local inset = db.Frame.Tooltip.Backdrop.Inset;
		local last = nil;
		local maxWidth, height = 0, (2 * inset) + frame.Drag:GetHeight() + 10;
		
		for i, v in ipairs(frame.Lines) do
			v:Show();
			v.Text:Show();
			v:ClearAllPoints();
			if not last then
				v:SetPoint("TOP", frame.Container, "TOP", 0, 0);
			else
				v:SetPoint("TOP", last, "BOTTOM", 0, 0);
			end
			v:SetPoint("LEFT", frame.Container, "LEFT", inset, 0);
			v:SetWidth(v.Text:GetStringWidth() + (2 * inset));
			v:SetHeight(v.Text:GetStringHeight());
			height = height + v:GetHeight();
			
			if v:GetWidth() > maxWidth then
				maxWidth = v:GetWidth();
			end
			last = v;
		end
		
		frame:SetWidth(maxWidth + (2 * inset));
		frame:SetHeight(height);
	end,
	UpdateData = function(frame)
		local lfgData, lfgMode = addon:UpdateQueueData();
		frame:ClearLines();
		
		if (not lfgMode.Mode or lfgMode.Mode == "abandonedInDungeon") and not lfgData.HaveData then
			frame:AddLine(L["You are not in an LFG queue"]);
		elseif lfgMode.Mode == "proposal" then
			frame:AddLine(L["Found a group"]);
			if lfgMode.SubMode == "accepted" then
				frame:AddLine(L["Waiting on group members"]);
			end
		elseif (lfgMode.Mode == "queued" or lfgMode.Mode == "listed" or lfgMode.Mode == "rolecheck") and not lfgData.HaveData then
			frame:AddLine(L["Acquiring LFG queue data..."]);
			if lfgData.QueueTime then
				self:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData.QueueTime)));
			end
		elseif lfgMode.Mode == "lfgparty" then
			frame:AddLine(L["You are in an LFG party"]);
		elseif lfgMode.Mode == "queued" and lfgData.HaveData then
			if db.Frame.Tooltip.Data.Type then
				frame:AddLine((L["Dungeon type"] .. ": %s"):format(lfgData.QueueName));
			end
			if db.Frame.Tooltip.Data.Tank then
				frame:AddLine((L["Tank queue times"] .. ": %s"):format(lfgData.TankWait == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData.TankWait)));
			end
			if db.Frame.Tooltip.Data.Healer then
				frame:AddLine((L["Healer queue times"] .. ": %s"):format(lfgData.HealerWait == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData.HealerWait)));
			end
			if db.Frame.Tooltip.Data.DPS then
				frame:AddLine((L["DPS queue times"] .. ": %s"):format(lfgData.DPSWait == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData.DPSWait)));
			end
			if db.Frame.Tooltip.Data.Self then
				frame:AddLine((L["Your queue time"] .. ": %s"):format(lfgData.MyWait == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData.MyWait)));
			end
			frame:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData.QueueTime)));
			if db.Frame.Tooltip.Data.Party then
				local haveTanks, haveHealers, haveDPS = lfgData.TotalTanks - lfgData.NeedTanks, lfgData.TotalHealers - lfgData.NeedHealers, lfgData.TotalDPS - lfgData.NeedDPS;
				local totalHave, totalNeed = haveTanks + haveHealers + haveDPS, lfgData.TotalTanks + lfgData.TotalHealers + lfgData.TotalDPS;
				if db.Frame.Tooltip.Data.Found.Tank then
					frame:AddLine((L["Tanks found"] .. ": %d/%d"):format(haveTanks, lfgData.TotalTanks));
				end
				if db.Frame.Tooltip.Data.Found.Healer then
					frame:AddLine((L["Healers found"] .. ": %d/%d"):format(haveHealers, lfgData.TotalHealers));
				end
				if db.Frame.Tooltip.Data.Found.DPS then
					frame:AddLine((L["DPS found"] .. ": %d/%d"):format(haveDPS, lfgData.TotalDPS));
				end
				if db.Frame.Tooltip.Data.Found.Total then
					frame:AddLine((L["Party members found"] .. ": %d/%d"):format(totalHave, totalNeed));
				end
			end
		else
			self:AddLine(L["Found an odd state."]);
			self:AddLine(L["Waiting on updated data from the server."]);
		end
		
		frame:UpdateSize();
	end,
};
